package parser

import "git.tebibyte.media/sashakoshka/fspl/lexer"
import "git.tebibyte.media/sashakoshka/fspl/errors"
import "git.tebibyte.media/sashakoshka/fspl/entity"

// Expression decision tree flow:
//
//   | +Ident =Variable
//   |        | 'true'       =LiteralBoolean
//   |        | 'false'      =LiteralBoolean
//   |        | 'nil'        =LiteralNil
//   |        | 'if'         =IfElse
//   |        | 'loop'       =Loop
//   |        | +Colon       =Declaration
//   |        | +DoubleColon =Call
//   |
//   | +LParen X
//   |         | +Star =LiteralArray
//   |         | +Dot  =LiteralStruct
//   |
//   | +LBracket | +Ident =Call
//   |           |        | 'break'  =Break
//   |           |        | 'return' =Return
//   |           | 
//   |           | +Dot +Expression =Dereference
//   |           |                  | +Expression =Subscript
//   |           | +Star =Operation
//   |           |
//   |           | +Symbol X
//   |                     | '\'      =Slice
//   |                     | '#'      =Length
//   |                     | '@'      =Reference
//   |                     | '~'      =ValueCast
//   |                     | '~~'     =BitCast
//   |                     | OPERATOR =Operation
//   |
//   | +LBrace =Block
//   | +Int    =LiteralInt
//   | +Float  =LiteralFloat
//   | +String =LiteralString
//
// Entities with a star have yet to be implemented

var descriptionExpression = "expression"
var startTokensExpression = []lexer.TokenKind {
	lexer.Ident,
	lexer.LParen,
	lexer.LBracket,
	lexer.LBrace,
	lexer.Int,
	lexer.Float,
	lexer.String,
}

func (this *Parser) parseExpression () (entity.Expression, error) {
	// Steps that this function takes to parse expressions:
	// 1. Run a decision tree to parse the expression.
	// 2. After that, test for infix operators (. and =). We will not need
	//    to consume more tokens to do this, as if they exist they will be
	//    the very next token AKA the current one after parsing the
	//    expression.
	// 3. Then, use the previously parsed expression and use it as the left
	//    side, and go on to parse the right side. Assignment (=) is the
	//    only infix expression where both left and right sides are
	//    expressions, so it gets to call parseExpression() and recurse. The
	//    implied associativity matters here.
	// 4. If not assignment, then this has to do with member access (the
	//    right side is not an expression), so we repeat from step 2 in order
	//    to chain member accesses and method calls. Step 2 will test
	//    whether or not we need to continue repeating afterward. The
	//    implied associativity matters here.
	// 5. Return the expression, whether it was directly parsed or is a tree
	//    of infix operator(s).

	// run decision tree
	expression, err := this.parseExpressionRoot()
	if err != nil { return nil, err }

	// test for infix operators
	for this.token.Is(lexer.Dot) || (this.token.Is(lexer.Symbol) && this.token.ValueIs("=")) {
		pos := this.pos()
	
		switch this.kind() {
		case lexer.Dot:
			// Dot: member access or method call
			// this control path must not return, staying within the
			// loop
			source := expression
			err := this.expectNextDesc (
				"member access or method call",
				lexer.Ident, lexer.LBracket)
			if err != nil { return nil, err }
			
			switch this.kind() {
			case lexer.Ident:
				// Ident: member access
				expression = &entity.MemberAccess {
					Position: pos.Union(this.pos()),
					Source:   source,
					Member:   this.value(),
				}
				this.next()
				
			case lexer.LBracket:
				// LBracket: method call
				expression, err = this.parseMethodCallCore(pos, source)
				if err != nil { return nil, err }
			}
			
		case lexer.Symbol:
			// Symbol '=': assignment
			// this control path must return, breaking out of the
			// loop.
			expression := &entity.Assignment {
				Position: pos,
				Location: expression,
			}

			this.next()
			value, err := this.parseExpression()
			if err != nil { return nil, err }
			expression.Value = value
			return expression, nil
		}
	}
	
	return expression, nil
}

func (this *Parser) parseExpressionRoot () (entity.Expression, error) {
	err := this.expectDesc(descriptionExpression, startTokensExpression...)
	if err != nil { return nil, err }
	
	switch this.kind() {
	case lexer.Ident:    return this.parseExpressionRootIdent()
	case lexer.LParen:   return this.parseExpressionRootLParen()
	case lexer.LBracket: return this.parseExpressionRootLBracket()
	case lexer.LBrace:   return this.parseBlock()
	case lexer.Int:      return this.parseLiteralInt()
	case lexer.Float:    return this.parseLiteralFloat()
	case lexer.String:   return this.parseLiteralString()
	}
	panic(this.bug())
}

func (this *Parser) parseExpressionRootIdent () (entity.Expression, error) {
	err := this.expect(lexer.Ident)
	if err != nil { return nil, err }
	name := this.value()
	pos  := this.pos()
	
	switch name {
	case "true", "false": return this.parseLiteralBoolean ()
	case "nil":           return this.parseLiteralNil()
	case "if":            return this.parseIfElse()
	case "loop":          return this.parseLoop()
	default:
		this.next()
		switch this.kind() {
		case lexer.Colon:
			// Colon: declaration
			return this.parseDeclarationCore(pos, name)
		case lexer.DoubleColon:
			// DoubleColon: call
			err := this.expectNext(lexer.LBracket)
			if err != nil { return nil, err }
			this.next()
			return this.parseCallCore(pos, name)
		default:
			// *: variable
			return &entity.Variable {
				Position: pos,
				Name:     name,
			}, nil
		}
	}
	panic(this.bug())
}

func (this *Parser) parseExpressionRootLParen() (entity.Expression, error) {
	err := this.expect(lexer.LParen)
	if err != nil { return nil, err }
	pos := this.pos()
	
	err = this.expectNext(lexer.Star, lexer.Dot)
	if err != nil { return nil, err }
	switch this.kind() {
	case lexer.Star: return this.parseLiteralArrayCore(pos)
	case lexer.Dot:  return this.parseLiteralStructCore(pos)
	}
	panic(this.bug())
}

func (this *Parser) parseExpressionRootLBracket () (entity.Expression, error) {
	err := this.expect(lexer.LBracket)
	if err != nil { return nil, err }
	pos := this.pos()
	
	err = this.expectNext(lexer.Ident, lexer.Dot, lexer.Star, lexer.Symbol)
	if err != nil { return nil, err }
	switch this.kind() {
	case lexer.Ident:  return this.parseExpressionRootLBracketIdent(pos)
	case lexer.Dot:    return this.parseDereferenceOrSubscriptCore(pos)
	case lexer.Star:   return this.parseOperationCore(pos)
	case lexer.Symbol: return this.parseExpressionRootLBracketSymbol(pos)
	}
	panic(this.bug())
}

func (this *Parser) parseExpressionRootLBracketIdent (pos errors.Position) (entity.Expression, error) {
	err := this.expect(lexer.Ident)
	if err != nil { return nil, err }
	name := this.value()

	switch name {
	case "break", "return": return this.parseReturnOrBreakCore(pos)
	default:                return this.parseCallCore(pos, "")
	}
	panic(this.bug())
}

func (this *Parser) parseDereferenceOrSubscriptCore (pos errors.Position) (entity.Expression, error) {
	err := this.expect(lexer.Dot)
	if err != nil { return nil, err }

	this.next()
	argument, err := this.parseExpression()
	if err != nil { return nil, err }

	err = this.expectDesc (
		"element offset expression or end of dereference",
		appendCopy(startTokensExpression, lexer.RBracket)...)
	if err != nil { return nil, err }
	
	if this.token.Is(lexer.RBracket) {
		// RBracket: dereference
		pos = pos.Union(this.pos())
		this.next()

		return &entity.Dereference {
			Position: pos,
			Pointer:  argument,
		}, nil
	} else {
		// startTokensExpression...: subscript
		offset, err := this.parseExpression()
		if err != nil { return nil, err }
		
		err = this.expect(lexer.RBracket)
		if err != nil { return nil, err }
		pos = pos.Union(this.pos())
		this.next()

		return &entity.Subscript {
			Position: pos,
			Slice:    argument,
			Offset:   offset,
		}, nil
	}
}

func (this *Parser) parseExpressionRootLBracketSymbol (pos errors.Position) (entity.Expression, error) {
	err := this.expectValue (
		lexer.Symbol,
		appendCopy(valuesOperator, "\\", "#", "@", "~", "~~")...)
	if err != nil { return nil, err }
	
	switch this.value() {
	case "\\": return this.parseSliceCore(pos)
	case "#":  return this.parseLengthCore(pos)
	case "@":  return this.parseReferenceCore(pos)
	case "~":  return this.parseValueOrBitCastCore(pos)
	case "~~": return this.parseValueOrBitCastCore(pos)
	default:   return this.parseOperationCore(pos)
	}
	panic(this.bug())
}

func (this *Parser) parseCallCore (pos errors.Position, module string) (*entity.Call, error) {
	err := this.expect(lexer.Ident)
	if err != nil { return nil, err }
	call := &entity.Call {
		Position: pos,
		Module:   module,
		Name:     this.value(),
	}
	this.next()

	for {
		err = this.expectDesc (
			"argument or end of call",
			appendCopy(startTokensExpression, lexer.RBracket)...)
		if err != nil { return nil, err }
		if this.token.Is(lexer.RBracket) { break }
		
		argument, err := this.parseExpression()
		if err != nil { return nil, err }
		call.Arguments = append(call.Arguments, argument)
	}
	call.Position = call.Position.Union(this.pos())
	this.next()

	return call, nil
}

func (this *Parser) parseMethodCallCore (pos errors.Position, source entity.Expression) (*entity.MethodCall, error) {
	err := this.expect(lexer.LBracket)
	if err != nil { return nil, err }
	err = this.expectNext(lexer.Ident)
	if err != nil { return nil, err }
	call := &entity.MethodCall {
		Source:   source,
		Position: pos,
		Name:     this.value(),
	}
	this.next()

	for {
		err = this.expectDesc (
			"argument or end of method call",
			appendCopy(startTokensExpression, lexer.RBracket)...)
		if err != nil { return nil, err }
		if this.token.Is(lexer.RBracket) { break }
		
		argument, err := this.parseExpression()
		if err != nil { return nil, err }
		call.Arguments = append(call.Arguments, argument)
	}
	call.Position = call.Position.Union(this.pos())
	this.next()

	return call, nil
}

func (this *Parser) parseReturnOrBreakCore (pos errors.Position) (entity.Expression, error) {
	err := this.expectValue(lexer.Ident, "break", "return")
	if err != nil { return nil, err }
	name := this.value()
	
	err = this.expectNextDesc (
		"expression or end of " + name,
		appendCopy(startTokensExpression, lexer.RBracket)...)
	if err != nil { return nil, err }
	var value entity.Expression
	if this.token.Is(startTokensExpression...) {
		// startTokensExpression...: value expression
		value, err = this.parseExpression()
		if err != nil { return nil, err }
	}

	err = this.expectDesc("end of " + name, lexer.RBracket)
	if err != nil { return nil, err }
	pos = pos.Union(this.pos())
	this.next()

	switch name {
	case "break":
		return &entity.Break {
			Position: pos,
			Value:    value,
		}, nil
	case "return":
		return &entity.Return {
			Position: pos,
			Value:    value,
		}, nil
	}
	panic(this.bug())
}

func (this *Parser) parseSliceCore (pos errors.Position) (*entity.Slice, error) {
	err := this.expectValue(lexer.Symbol , "\\")
	if err != nil { return nil, err }
	slice := &entity.Slice {
		Position: pos,
	}
	
	err = this.expectNextDesc(descriptionExpression, startTokensExpression...)
	if err != nil { return nil, err }
	slice.Slice, err = this.parseExpression()

	err = this.expectDesc (
		"offset expression or '/'",
		appendCopy(startTokensExpression, lexer.Symbol)...)
	if err != nil { return nil, err }
	if this.token.Is(startTokensExpression...) {
		// startTokensExpression...: starting offset
		slice.Start, err = this.parseExpression()
		if err != nil { return nil, err }
	}

	err = this.expectValue(lexer.Symbol, "/")
	if err != nil { return nil, err }
	
	this.expectNextDesc (
		"offset expression or operation end",
		appendCopy(startTokensExpression, lexer.RBracket)...)
	if this.token.Is(startTokensExpression...) {
		// startTokensExpression...: ending offset
		slice.End, err = this.parseExpression()
		if err != nil { return nil, err }
	}

	err = this.expect(lexer.RBracket)
	if err != nil { return nil, err }
	slice.Position = slice.Position.Union(this.pos())
	this.next()

	return slice, nil
}

func (this *Parser) parseLengthCore (pos errors.Position) (*entity.Length, error) {
	err := this.expectValue(lexer.Symbol , "#")
	if err != nil { return nil, err }
	length := &entity.Length {
		Position: pos,
	}
	
	this.next()
	length.Slice, err = this.parseExpression()
	if err != nil { return nil, err }

	err = this.expect(lexer.RBracket)
	if err != nil { return nil, err }
	length.Position = length.Position.Union(this.pos())
	this.next()

	return length, nil
}

func (this *Parser) parseReferenceCore (pos errors.Position) (*entity.Reference, error) {
	err := this.expectValue(lexer.Symbol , "@")
	if err != nil { return nil, err }
	reference := &entity.Reference {
		Position: pos,
	}
	
	this.next()
	reference.Value, err = this.parseExpression()
	if err != nil { return nil, err }

	err = this.expect(lexer.RBracket)
	if err != nil { return nil, err }
	reference.Position = reference.Position.Union(this.pos())
	this.next()

	return reference, nil
}

func (this *Parser) parseValueOrBitCastCore (pos errors.Position) (entity.Expression, error) {
	err := this.expectValue(lexer.Symbol , "~", "~~")
	if err != nil { return nil, err }
	tokValue := this.value()

	this.next()
	ty, err := this.parseType()
	if err != nil { return nil, err }
	
	err = this.expectDesc(descriptionExpression, startTokensExpression...)
	if err != nil { return nil, err }
	value, err := this.parseExpression()

	err = this.expect(lexer.RBracket)
	if err != nil { return nil, err }
	pos = pos.Union(this.pos())
	this.next()

	switch tokValue {
	case "~":
		return &entity.ValueCast {
			Position: pos,
			Ty:       ty,
			Value:    value,
		}, nil
	case "~~":
		return &entity.BitCast {
			Position: pos,
			Ty:       ty,
			Value:    value,
		}, nil
	}
	panic(this.bug())
}

var valuesOperator = []string {
	"++", "+", "--", "-", "*", "/", "%", "!!", "||", "&&", "^^",
	"!", "|", "&", "^", "<<", ">>", "<", ">", "<=", ">=", "=",
}

func (this *Parser) parseOperationCore (pos errors.Position) (*entity.Operation, error) {
	// TODO: maybe come up with a more elegant way of writing this?
	// possibly make a verion of expectValue that matches a list of kinds
	// and a list of values. could also make a version that accepts any kind
	// of token, and always supplement that with a normal expect.
	if this.token.Is(lexer.Star) {
		err := this.expectValue(lexer.Star, valuesOperator...)
		if err != nil { return nil, err }
	} else {
		err := this.expectValue(lexer.Symbol, valuesOperator...)
		if err != nil { return nil, err }
	}
	
	operation := &entity.Operation {
		Position: pos,
		Operator: entity.OperatorFromString(this.value()),
	}
	this.next()

	for {
		err := this.expectDesc (
			"expression or operation end",
			appendCopy(startTokensExpression, lexer.RBracket)...)
		if err != nil { return nil, err }
		if this.token.Is(lexer.RBracket) { break }
		
		argument, err := this.parseExpression()
		if err != nil { return nil, err }
		operation.Arguments = append(operation.Arguments, argument)
	}
	operation.Position = operation.Position.Union(this.pos())
	this.next()

	return operation, nil
}

func (this *Parser) parseBlock () (*entity.Block, error) {
	err := this.expectDesc("Block", lexer.LBrace)
	if err != nil { return nil, err }
	block := &entity.Block {
		Position: this.pos(),
	}
	this.next()

	for {
		err := this.expectDesc (
			descriptionExpression,
			appendCopy(startTokensExpression, lexer.RBrace)...)
		if err != nil { return nil, err }
		if this.kind() == lexer.RBrace{ break }

		step, err := this.parseExpression()
		if err != nil { return nil, err }
		block.Steps = append(block.Steps, step)
	}
	block.Position = block.Position.Union(this.pos())
	this.next()
	
	return block, nil
}

var descriptionDeclaration = "declaration"
var startTokensDeclaration = []lexer.TokenKind { lexer.Ident }

func (this *Parser) parseDeclaration () (*entity.Declaration, error) {
	err := this.expectDesc(descriptionDeclaration, startTokensDeclaration...)
	if err != nil { return nil, err }
	name := this.value()
	pos  := this.pos()
	this.next()
	return this.parseDeclarationCore(pos, name)
}

func (this *Parser) parseDeclarationCore (pos errors.Position, name string) (*entity.Declaration, error) {
	err := this.expect(lexer.Colon)
	if err != nil { return nil, err }
	this.next()
	ty, err := this.parseType()
	if err != nil { return nil, err }
	
	return &entity.Declaration {
		Position: pos,
		Name:     name,
		Ty:       ty,
	}, nil
}

func (this *Parser) parseIfElse () (*entity.IfElse, error) {
	err := this.expectValue(lexer.Ident, "if")
	if err != nil { return nil, err }
	ifElse := &entity.IfElse {
		Position: this.pos(),
	}

	err = this.expectNextDesc("condition", startTokensExpression...)
	if err != nil { return nil, err }
	ifElse.Condition, err = this.parseExpression()
	if err != nil { return nil, err }

	err = this.expectValue(lexer.Ident, "then")
	if err != nil { return nil, err }

	this.expectNextDesc("branch expression", startTokensExpression...)
	ifElse.True, err = this.parseExpression()
	if err != nil { return nil, err }

	if this.token.Is(lexer.Ident) && this.token.ValueIs("else") {
		// Ident "else": false branch expression
		this.expectNextDesc("Branch expression", startTokensExpression...)
		ifElse.False, err = this.parseExpression()
		if err != nil { return nil, err }
	}
	
	return ifElse, nil
}

func (this *Parser) parseLoop () (*entity.Loop, error) {
	err := this.expectValue(lexer.Ident, "loop")
	if err != nil { return nil, err }
	pos := this.pos()

	this.next()
	body, err := this.parseExpression()
	if err != nil { return nil, err }

	return &entity.Loop {
		Position: pos,
		Body:     body,
	}, nil
}
