package parser

import "git.tebibyte.media/sashakoshka/fspl/lexer"
import "git.tebibyte.media/sashakoshka/fspl/errors"
import "git.tebibyte.media/sashakoshka/fspl/entity"

var descriptionTopLevel = "typedef, function, or method"
var startTokensTopLevel = []lexer.TokenKind {
	lexer.Symbol,
	lexer.LBracket,
	lexer.TypeIdent,
	lexer.EOF,	
}

func (this *Parser) parseTopLevel () error {
	err := this.expectDesc (
		descriptionTopLevel,
		startTokensTopLevel...)
	if err != nil { return err }
	if this.token.EOF() { return nil }
	pos := this.pos()
	
	access := entity.AccessPrivate
	if this.token.Kind == lexer.Symbol {
		access, err = this.parseAccess()

		err = this.expectDesc (
			descriptionTopLevel,
			lexer.Symbol,
			lexer.LBracket,
			lexer.Ident,
			lexer.TypeIdent)
		if err != nil { return err }
	}

	// LBracket: Function
	if this.token.Is(lexer.LBracket) {
		function, err := this.parseFunctionCore(pos, access)
		if err != nil { return err }
		this.tree.AddDeclaration(function)
		return nil
	}

	// TypeIdent: Method, or Typedef
	typeName := this.token.Value
	err = this.expectNext(lexer.Dot, lexer.Colon)
	if err != nil { return err }

	switch this.token.Kind {
	case lexer.Dot:
		// Dot: Method
		err = this.expectNextDesc(descriptionSignature, startTokensSignature...)
		if err != nil { return err }
		method, err := this.parseMethodCore(pos, access, typeName)
		if err != nil { return err }
		this.tree.AddDeclaration(method)
	case lexer.Colon:
		// Colon: Typedef
		err = this.expectNextDesc(descriptionType, startTokensType...)
		if err != nil { return err }
		typedef, err := this.parseTypedefCore(pos, access, typeName)
		if err != nil { return err }
		this.tree.AddDeclaration(typedef)
	default: this.bug()
	}
	return nil
}

func (this *Parser) parseAccess () (entity.Access, error) {
	err := this.expectValueDesc (
		"Access control specifier",
		lexer.Symbol, "-", "~", "+")
	if err != nil { return 0, err }
	defer this.next()
	
	switch this.token.Value {
	case "-": return entity.AccessPrivate,    nil
	case "~": return entity.AccessRestricted, nil
	case "+": return entity.AccessPublic,     nil
	default: panic(this.bug())
	}
}

func (this *Parser) parseFunctionCore (pos errors.Position, access entity.Access) (*entity.Function, error) {
	signature, err := this.parseSignature()
	if err != nil { return nil, err }

	function := &entity.Function {
		Position:  pos.Union(signature.Position),
		Acc:       access,
		Signature: signature,
	}

	err = this.expectDesc (
		"function body, link name, or top-level declaration",
		appendCopy(startTokensTopLevel, lexer.Symbol, lexer.String)...)
	if err != nil { return nil, err }

	if this.token.Is(lexer.String) {
		// String: Link name
		function.LinkName = this.token.Value
		err = this.expectNextDesc (
			"Function body, " + descriptionTopLevel,
			appendCopy(startTokensTopLevel, lexer.Symbol)...)
		if err != nil { return nil, err }
	}

	if !(this.token.Is(lexer.Symbol) && this.token.ValueIs("=")) {
		// function has no body and is finished,
		// move on to next top level declaration or graceful EOF.
		return function, nil
	}

	// Symbol '=': Function body
	err = this.expectValueDesc("function body", lexer.Symbol, "=")
	if err != nil { return nil, err }

	// Expression
	err = this.expectNextDesc(descriptionExpression, startTokensExpression...)
	if err != nil { return nil, err }
	bodyExpression, err := this.parseExpression()
	if err != nil { return nil, err }
	function.Body = bodyExpression
	return function, nil
}

func (this *Parser) parseMethodCore (pos errors.Position, access entity.Access, typeName string) (*entity.Method, error) {
	function, err := this.parseFunctionCore(pos, access)
	if err != nil { return nil, err }
	return &entity.Method {
		Position:  function.Position,
		Acc:       function.Acc,
		TypeName:  typeName,
		Signature: function.Signature,
		LinkName:  function.LinkName,
		Body:      function.Body,
	}, nil
}

func (this *Parser) parseTypedefCore (pos errors.Position, access entity.Access, typeName string) (*entity.Typedef, error) {
	pos = pos.Union(this.pos())
	ty, err := this.parseType()
	if err != nil { return nil, err }
	return &entity.Typedef {
		Position: pos,
		Acc:      access,
		Name:     typeName,
		Type:     ty,
	}, nil
}
